from finder.base import Finder
from itertools import chain
from pathlib import Path
from typing import List

import logging
import re


logger = logging.getLogger(__name__)


class FileFinder(Finder):
    """An interface for finding files. It combines best of both worlds glob/rglob (speed) and regex (flexibility):
    Mandatory arguments:
        'extension' (str)                       To filter files by extension.
        'single_start_dir' (pathlib.Path)       Either use single_start_dir or multi_start_dir. The search starts here.
        'multi_start_dir' (List[pathlib.Path])  Either use single_start_dir or multi_start_dir. The search starts
                                                from multiple start locations (one after another).
    Optional arguments:
        'limit_depth' (bool)                    To limit the recursive search depth.
        'depth' (int)                           To limit the recursive search depth.
        'filename_regex' (str)                  A regex to find filename e.g. "^[0-9]{8}_HistTags$"

    Example:
        single_start_dir    = pathlib.Path('this_is_my_start_dir')
        limit_depth         = True
        depth               = 0  # <-- 0 so do not search in possible subdirs
        filename_regex      = "^[0-9]{8}_HistTags$"
        extension           = '.csv'

        file_finder = FileFinder(
                            single_start_dir=single_start_dir,
                            extension=extension,
                            limit_depth=True,
                            depth=depth,
                            filename_regex=filename_regex
                            )

        paths = file_finder.paths
        paths_empty_files = file_finder.paths_empty_file

    """

    ALL_EXTENSIONS = "*"
    EXTENTION_CHOICES = (
        ".jpg",
        ".png",
        ".txt",
        ".xml",
        ".csv",
        ".xlsx",
        ".pdf",
        ".h5",
        ".nc",
        ".zip",
    )

    def __init__(self, extension: str, filename_regex: str = None, *args, **kwargs):
        self.extension = extension if extension else ""
        self.filename_regex = filename_regex
        self.validate_filefinder_constructor()
        self._paths = None
        self._paths_empty_file = None
        super().__init__(*args, **kwargs)

    def validate_filefinder_constructor(self) -> None:
        if not isinstance(self.extension, str):
            raise AssertionError("extension must either be None or a str")
        if (
            self.extension not in self.EXTENTION_CHOICES
            and self.extension != self.ALL_EXTENSIONS
        ):
            raise AssertionError(
                f"extension {self.extension} must either be None or a str in {self.EXTENTION_CHOICES}"
            )

        # filename_regex is optional!!
        if self.filename_regex and not isinstance(self.filename_regex, str):
            raise AssertionError("filename_regex must be a str")

    def _get_paths_from_single_dir(self, single_dir: Path) -> List[Path]:

        if self.limit_depth:
            file_paths_generator = chain()
            for _depth_n, glob_pattern in self.DEPTH_MAPPER.items():
                if _depth_n > self.depth:
                    break

                if self.filename_regex:
                    only_files_generator = (
                        _path
                        for _path in single_dir.glob(glob_pattern + self.extension)
                        if _path.is_file()
                        and re.match(pattern=self.filename_regex, string=_path.stem)
                    )
                else:
                    only_files_generator = (
                        _path
                        for _path in single_dir.glob(glob_pattern + self.extension)
                        if _path.is_file()
                    )
                # merge generators into one
                file_paths_generator = chain(file_paths_generator, only_files_generator)
            logger.debug("convert generator to list, this may take a while")
            return [x for x in file_paths_generator]

        if not self.limit_depth:
            # note we use rglob (recursive search all subdirs)
            if self.filename_regex:
                file_paths_generator = (
                    _path
                    for _path in single_dir.rglob(f"*{self.extension}")
                    if _path.is_file()
                    and re.match(pattern=self.filename_regex, string=_path.stem)
                )
            else:
                file_paths_generator = (
                    _path
                    for _path in single_dir.rglob(f"*{self.extension}")
                    if _path.is_file()
                )
            logger.debug("convert generator to list, this may take a while")
            return [x for x in file_paths_generator]

    def _get_paths_from_multi_dir(self) -> List[Path]:
        nested_lists_with_paths = [
            self._get_paths_from_single_dir(single_dir=_dir_path)
            for _dir_path in self.multi_start_dir
        ]
        paths_from_multi_dir = [
            item for sublist in nested_lists_with_paths for item in sublist
        ]
        return paths_from_multi_dir if paths_from_multi_dir else []

    @property
    def paths(self) -> List[Path]:
        if self._paths or self._paths == []:
            return self._paths
        if self.single_start_dir:
            self._paths = self._get_paths_from_single_dir(
                single_dir=self.single_start_dir
            )
        elif self.multi_start_dir:
            self._paths = self._get_paths_from_multi_dir()
        return self._paths

    @property
    def paths_empty_file(self) -> List[Path]:
        """ A selection of self.paths of files that are empty (filesize=0kb). """
        if self._paths_empty_file or self._paths_empty_file == []:
            return self._paths_empty_file
        self._paths_empty_file = [
            _path for _path in self.paths if _path.stat().st_size == 0
        ]
        return self._paths_empty_file



from finder.base import Finder
from itertools import chain
from pathlib import Path
from typing import List

import logging
import re


logger = logging.getLogger(__name__)


class DirFinder(Finder):
    def __init__(
        self,
        dirname_regex: str = None,
        exclude_empty_dirs: bool = False,
        *args,
        **kwargs,
    ):
        self.dirname_regex = dirname_regex
        self.exclude_empty_dirs = exclude_empty_dirs
        self._paths = None
        self._paths_empty_dir = None
        self.validate_dirfinder_constructor()
        super().__init__(*args, **kwargs)

    def validate_dirfinder_constructor(self):
        # filename_regex is optional!!
        if self.dirname_regex and not isinstance(self.dirname_regex, str):
            raise AssertionError("dirname_regex must be a str")

    def _is_dir_path_regex_match(self, _path: Path) -> bool:
        return _path.is_dir() and re.match(
            pattern=self.dirname_regex, string=_path.stem
        )

    def _get_paths_from_single_dir(self, single_dir: Path) -> List[Path]:
        if self.limit_depth:
            # When we get all recursive paths with rglob('*') and then evaluate them may result in
            # potentially a lot of unnecessary work. Solution below is 'do it depth-by-depth': if e.g.
            # self.depth=2, then we get first all paths of depth=0, then depth=1, and then depth=2.
            # First, create an empty generator in which we will merge one generator per depth
            dir_paths_generator = chain()
            for _depth_n, glob_pattern in self.DEPTH_MAPPER.items():
                if _depth_n > self.depth:
                    break
                if self.dirname_regex:
                    only_dirs_generator = (
                        _path
                        for _path in single_dir.glob(glob_pattern)
                        if _path.is_dir()
                        and re.match(pattern=self.dirname_regex, string=_path.stem)
                    )
                else:
                    only_dirs_generator = (
                        _path
                        for _path in single_dir.glob(glob_pattern)
                        if _path.is_dir()
                    )
                # merge generators into one
                dir_paths_generator = chain(dir_paths_generator, only_dirs_generator)
            logger.debug("convert generator to list, this may take a while")
            return [x for x in dir_paths_generator]

        if not self.limit_depth:
            # note we use rglob (recursive search all subdirs)
            if self.dirname_regex:
                dir_paths_generator = (
                    _path
                    for _path in single_dir.rglob("*")
                    if _path.is_dir()
                    and re.match(pattern=self.dirname_regex, string=_path.stem)
                )

            else:
                dir_paths_generator = (
                    _path for _path in single_dir.rglob("*") if _path.is_dir()
                )
            logger.debug("convert generator to list, this may take a while")
            return [x for x in dir_paths_generator]

    def _get_paths_from_multi_dir(self) -> List[Path]:
        nested_lists_with_paths = [
            self._get_paths_from_single_dir(single_dir=_dir_path)
            for _dir_path in self.multi_start_dir
        ]
        paths_from_multi_dir = [
            item for sublist in nested_lists_with_paths for item in sublist
        ]
        return list(set(paths_from_multi_dir)) if paths_from_multi_dir else []

    @property
    def paths(self) -> List[Path]:
        if self._paths or self._paths == []:
            return self._paths

        # single dir
        if self.single_start_dir:
            self._paths = self._get_paths_from_single_dir(
                single_dir=self.single_start_dir
            )
            if self.exclude_empty_dirs:
                self._paths = [_path for _path in self._paths if any(_path.iterdir())]
            return self._paths

        # multi dir
        self._paths = self._get_paths_from_multi_dir()
        if self.exclude_empty_dirs:
            self._paths = [_path for _path in self._paths if any(_path.iterdir())]
        return self._paths

    @property
    def paths_empty_dir(self) -> List[Path]:
        """ A selection of self.paths of dirs that hold no files. """
        if self._paths_empty_dir or self._paths_empty_dir == []:
            return self._paths_empty_dir
        elif self.exclude_empty_dirs:
            logger.info(
                f"paths_empty_dir is [] as search was done with exclude_empty_dirs=True"
            )
            self._paths_empty_dir = []
            return self._paths_empty_dir
        self._paths_empty_dir = [
            _path for _path in self.paths if not any(_path.iterdir())
        ]
        return self._paths_empty_dir



from pathlib import Path
from typing import List

import logging


logger = logging.getLogger(__name__)


class Finder:

    DEPTH_MAPPER = {
        0: "*",  # search only start_dir
        1: "*/*",  # search start_dir + its subdirs
        2: "*/*/*",  # etc..
        3: "*/*/*/*",
        4: "*/*/*/*/*",
        5: "*/*/*/*/*/*",
        6: "*/*/*/*/*/*/*",
    }

    def __init__(
        self,
        single_start_dir: Path = None,
        multi_start_dir: List[Path] = None,
        limit_depth: bool = True,
        depth: int = 0,
    ):
        """
        :param single_start_dir: Path:      One single path (=directory) from where the search starts.
        :param multi_start_dir: List[Path]: You can start your search from multiple paths directories.
        :param limit_depth: bool:           if False, then search all subdirs. If True then limit search to 'depth'.
        :param depth: int:                  nr of directories deep (recursively). E.g. depth=1? then only search in
                                            dir, all subdirs of dir
        Either choose single_start_dir or multi_start_dir
        """
        self.single_start_dir = single_start_dir
        self.multi_start_dir = multi_start_dir
        self.limit_depth = limit_depth
        self.depth = depth
        self.validate_finder_constructor()

    def validate_finder_constructor(self):
        # validate single_start_dir + multi_start_dir
        if (self.single_start_dir and self.multi_start_dir) or (
            not self.single_start_dir and not self.multi_start_dir
        ):
            raise AssertionError(f"use either single_start_dir or multi_start_dir")
        if self.single_start_dir:
            assert isinstance(
                self.single_start_dir, Path
            ), "single_start_dir must be a pathlib.Path"
            assert (
                self.single_start_dir.is_dir()
            ), f"single_start_dir {self.single_start_dir} does not exist"
        elif self.multi_start_dir:
            assert isinstance(
                self.multi_start_dir, list
            ), "multi_start_dir must be a list (with pathlib.Path)"
            none_path_objects = [
                x for x in self.multi_start_dir if not isinstance(x, Path)
            ]
            if none_path_objects:
                msg = "not all elements in multi_start_dir are of type pathlib.Path"
                if (
                    len(none_path_objects) < 4
                ):  # 4 is a bit random (just do not return too many paths..)
                    raise AssertionError(f"{msg} : {none_path_objects}")
                raise AssertionError(msg)
            none_existing_dirs = [x for x in self.multi_start_dir if not x.is_dir()]
            if none_existing_dirs:
                msg = "not all elements in multi_start_dir are existing directories"
                if (
                    len(none_existing_dirs) < 4
                ):  # 4 is a bit random (just do not return too many paths..)
                    raise AssertionError(f"{msg}: {none_existing_dirs}")
                raise AssertionError(msg)

        # validate depth + limit_depth
        assert isinstance(self.limit_depth, bool), f"limit_depth must be a bool"
        if self.depth and not self.limit_depth:
            raise AssertionError(
                f"depth={self.depth} is only possible with limit_depth=True"
            )
        if not self.limit_depth:
            return
        max_allowed_depth = max(self.DEPTH_MAPPER.keys())
        if not isinstance(self.depth, int) or not 0 <= self.depth <= max_allowed_depth:
            raise AssertionError(
                f"depth {self.depth} must be a int and in range: 0 <= depth <= {max_allowed_depth}"
            )
        logger.debug(
            f"search recursively with limit_depth=True with depth={self.depth}"
        )

    def _depth_to_startdir(self, path: Path, start_dir: Path) -> int:
        """Calculate the nr of parts (between slashes) in a relative path to start_dir.
        Returns 0 if path == start_dir,
        Returns 1 if path is a subdir of start_dir,
        etc..
        """
        try:
            parts = path.relative_to(start_dir).parts
            return len(parts) - 1
        except TypeError as err:
            raise AssertionError(
                f"path {path} could not related to start_dir {start_dir}, err={err}"
            )



from os import path
from setuptools import find_packages
from setuptools import setup


# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

version = "1.3"

install_requires = [
    "pathlib",
    "typing",
]

tests_require = [
    "pytest",
]

setup(
    name="path_finder",
    packages=find_packages(include=["finder", "finder.*"]),
    # packages=["finder"],
    # package_dir={"finder": "."},
    version=version,
    license="MIT",
    description="interface for finding directories and files",
    long_description_content_type="text/markdown",
    long_description=long_description,
    author="Renier Kramer",
    author_email="renier.kramer@hdsr.nl",
    url="https://github.com/hdsr-mid/path_finder",
    download_url=f"https://github.com/hdsr-mid/path_finder/archive/v{version}.tar.gz",
    keywords=["interface", "path", "directory", "filename", "glob", "regex", "find"],
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require={"test": tests_require},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)
